@@ -7,7 +7,7 @@
 // Copyright 2006-2008 Julien Rebetez <julienr@svn.gnome.org>
 // Copyright 2007-2012, 2015-2020 Albert Astals Cid <aacid@kde.org>
 // Copyright 2007-2008, 2011 Carlos Garcia Campos <carlosgc@gnome.org>
-// Copyright 2007, 2013, 2016 Adrian Johnson <ajohnson@redneon.com>
+// Copyright 2007, 2013, 2016, 2019 Adrian Johnson <ajohnson@redneon.com>
 // Copyright 2007 Iñigo Martínez <inigomartinez@gmail.com>
 // Copyright 2008, 2011 Pino Toscano <pino@kde.org>
 // Copyright 2008 Michael Vrable <mvrable@cs.ucsd.edu>
@@ -28,6 +28,8 @@
 // Copyright 2019 Tomoyuki Kubota <himajin100000@gmail.com>
 // Copyright 2019 João Netto <joaonetto901@gmail.com>
 // Copyright 2020 Marek Kasik <mkasik@redhat.com>
+// Copyright 2020 Thorsten Behrens <Thorsten.Behrens@CIB.de>
+// Copyright 2020 Klarälvdalens Datakonsult AB, a KDAB Group company, <info@kdab.com>. Work sponsored by Technische Universität Dresden
 //
 //========================================================================
 
@@ -40,8 +42,10 @@
 #include <cstring>
 #include <cctype>
 #include "goo/gmem.h"
+#include "goo/gfile.h"
 #include "goo/GooString.h"
 #include "Error.h"
+#include "ErrorCodes.h"
 #include "Object.h"
 #include "Array.h"
 #include "Dict.h"
@@ -53,11 +57,13 @@
 #    include "SignatureHandler.h"
 #endif
 #include "SignatureInfo.h"
+#include "CertificateInfo.h"
 #include "XRef.h"
 #include "PDFDocEncoding.h"
 #include "Annot.h"
 #include "Link.h"
 #include "Lexer.h"
+#include "Parser.h"
 
 // return a newly allocated char* containing an UTF16BE string of size length
 char *pdfDocEncodingToUTF16(const std::string &orig, int *length)
@@ -488,6 +494,241 @@ SignatureInfo *FormWidgetSignature::validateSignature(bool doVerifyCert, bool fo
     return static_cast<FormFieldSignature *>(field)->validateSignature(doVerifyCert, forceRevalidation, validationTime);
 }
 
+#ifdef ENABLE_NSS3
+// update hash with the specified range of data from the file
+static void hashFileRange(FILE *f, SignatureHandler *handler, Goffset start, Goffset end)
+{
+    const int BUF_SIZE = 65536;
+
+    unsigned char *buf = new unsigned char[BUF_SIZE];
+
+    while (start < end) {
+        Gfseek(f, start, SEEK_SET);
+        int len = BUF_SIZE;
+        if (end - start < len)
+            len = end - start;
+        fread(buf, len, 1, f);
+        handler->updateHash(buf, len);
+        start += len;
+    }
+    delete[] buf;
+}
+#endif
+
+bool FormWidgetSignature::signDocument(const char *saveFilename, const char *certNickname, const char *digestName, const char *password, const char *reason)
+{
+#ifdef ENABLE_NSS3
+    if (!certNickname) {
+        fprintf(stderr, "signDocument: Empty nickname\n");
+        return false;
+    }
+
+    // calculate a signature over tmp_buffer with the certificate to get its size
+    unsigned char tmp_buffer[4];
+    memcpy(tmp_buffer, "PDF", 4);
+    SignatureHandler sigHandler(certNickname, SEC_OID_SHA256);
+    sigHandler.updateHash(tmp_buffer, 4);
+    const std::unique_ptr<GooString> tmpSignature = sigHandler.signDetached(password);
+    if (!tmpSignature)
+        return false;
+
+    FormFieldSignature *signatureField = static_cast<FormFieldSignature *>(field);
+    std::unique_ptr<X509CertificateInfo> certInfo = sigHandler.getCertificateInfo();
+    const std::string signerName = certInfo->getSubjectInfo().commonName;
+    signatureField->setCertificateInfo(certInfo);
+    updateWidgetAppearance(); // add visible signing info to appearance
+
+    GooString gReason(reason ? reason : "");
+    Object vObj(new Dict(xref));
+    Ref vref = xref->addIndirectObject(&vObj);
+    if (!createSignature(vObj, vref, GooString(signerName), gReason, tmpSignature.get())) {
+        return false;
+    }
+
+    // Incremental save to avoid breaking any existing signatures
+    GooString *fname = new GooString(saveFilename);
+    if (doc->saveAs(fname, writeForceIncremental) != errNone) {
+        fprintf(stderr, "signDocument: error saving to file \"%s\"\n", saveFilename);
+        return false;
+    }
+
+    // Get start/end offset of signature object in the saved PDF
+    Goffset objStart, objEnd;
+    if (!getObjectStartEnd(fname, vref.num, &objStart, &objEnd)) {
+        fprintf(stderr, "signDocument: unable to get signature object offsets\n");
+    }
+
+    // Update byte range of signature in the saved PDF
+    Goffset sigStart, sigEnd, fileSize;
+    FILE *file = openFile(saveFilename, "r+b");
+    if (!updateOffsets(file, objStart, objEnd, &sigStart, &sigEnd, &fileSize)) {
+        fprintf(stderr, "signDocument: unable update byte range\n");
+        return false;
+    }
+
+    // compute hash of byte ranges
+    sigHandler.restartHash();
+    hashFileRange(file, &sigHandler, 0LL, sigStart);
+    hashFileRange(file, &sigHandler, sigEnd, fileSize);
+
+    // and sign it
+    const std::unique_ptr<GooString> signature = sigHandler.signDetached(password);
+    if (!signature)
+        return false;
+
+    // write signature to saved file
+    if (!updateSignature(file, sigStart, sigEnd, signature.get())) {
+        fprintf(stderr, "signDocument: unable update signature\n");
+        return false;
+    }
+    signatureField->setSignature(*signature);
+
+    fclose(file);
+
+    return true;
+#else
+    return false;
+#endif
+}
+
+// Get start and end file position of objNum in the PDF named filename.
+bool FormWidgetSignature::getObjectStartEnd(GooString *filename, int objNum, Goffset *objStart, Goffset *objEnd)
+{
+    PDFDoc newDoc(filename);
+    if (!newDoc.isOk())
+        return false;
+
+    XRef *newXref = newDoc.getXRef();
+    XRefEntry *entry = newXref->getEntry(objNum);
+    if (entry->type != xrefEntryUncompressed)
+        return false;
+
+    *objStart = entry->offset;
+    newXref->fetch(objNum, entry->gen, 0, objEnd);
+    return true;
+}
+
+// find next offset containing the dummy offset '9999999999' and overwrite with offset
+static char *setNextOffset(char *start, Goffset offset)
+{
+    char buf[50];
+    sprintf(buf, "%lld", offset);
+    strcat(buf, "                  "); // add some padding
+
+    char *p = strstr(start, "9999999999");
+    if (p) {
+        strncpy(p, buf, 10); // overwrite exact size.
+        p += 10;
+    } else {
+        return nullptr;
+    }
+    return p;
+}
+
+// Updates the ByteRange array of the signature object in the file.
+// Returns start/end of signature string and file size.
+bool FormWidgetSignature::updateOffsets(FILE *f, Goffset objStart, Goffset objEnd, Goffset *sigStart, Goffset *sigEnd, Goffset *fileSize)
+{
+    Gfseek(f, 0, SEEK_END);
+    *fileSize = Gftell(f);
+
+    if (objEnd > *fileSize)
+        objEnd = *fileSize;
+
+    // sanity check object offsets
+    if (objEnd <= objStart || (objEnd - objStart >= INT_MAX)) {
+        return false;
+    }
+
+    int bufSize = static_cast<int>(objEnd - objStart);
+    Gfseek(f, objStart, SEEK_SET);
+    char *buf = (char *)gmalloc_checkoverflow(bufSize + 1);
+    if (!buf)
+        return false;
+    fread(buf, bufSize, 1, f);
+    buf[bufSize] = 0; // prevent string functions from searching past the end
+
+    // search for the Contents field which contains the signature
+    // which always must start with hex digits 308
+    *sigStart = -1;
+    *sigEnd = -1;
+    for (int i = 0; i < bufSize - 14; i++) {
+        if (buf[i] == '/' && strncmp(&buf[i], "/Contents <308", 14) == 0) {
+            *sigStart = objStart + i + 10;
+            char *p = strchr(&buf[i], '>');
+            if (p)
+                *sigEnd = objStart + (p - buf) + 1;
+            break;
+        }
+    }
+
+    if (*sigStart == -1 || *sigEnd == -1)
+        return false;
+
+    // Search for ByteRange array and update offsets
+    for (int i = 0; i < bufSize - 10; i++) {
+        if (buf[i] == '/' && strncmp(&buf[i], "/ByteRange", 10) == 0) {
+            // update range
+            char *p = setNextOffset(&buf[i], *sigStart);
+            if (!p)
+                return false;
+            p = setNextOffset(p, *sigEnd);
+            if (!p)
+                return false;
+            p = setNextOffset(p, *fileSize - *sigEnd);
+            if (!p)
+                return false;
+            break;
+        }
+    }
+
+    // write buffer back to disk
+    Gfseek(f, objStart, SEEK_SET);
+    fwrite(buf, bufSize, 1, f);
+    free(buf);
+    return true;
+}
+
+// Overwrite signature string in the file with new signature
+bool FormWidgetSignature::updateSignature(FILE *f, Goffset sigStart, Goffset sigEnd, const GooString *signature)
+{
+    if (signature->getLength() * 2 + 2 != sigEnd - sigStart)
+        return false;
+
+    Gfseek(f, sigStart, SEEK_SET);
+    const char *c = signature->c_str();
+    fprintf(f, "<");
+    for (int i = 0; i < signature->getLength(); i++) {
+        unsigned char value = *(c + i) & 0x000000ff;
+        fprintf(f, "%2.2x", value);
+    }
+    fprintf(f, "> ");
+    return true;
+}
+
+bool FormWidgetSignature::createSignature(Object &vObj, Ref vRef, const GooString &name, const GooString &reason, const GooString *signature)
+{
+    vObj.dictAdd("Type", Object(objName, "Sig"));
+    vObj.dictAdd("Filter", Object(objName, "Adobe.PPKLite"));
+    vObj.dictAdd("SubFilter", Object(objName, "adbe.pkcs7.detached"));
+    vObj.dictAdd("Name", Object(name.copy()));
+    GooString *date = timeToDateString(nullptr);
+    vObj.dictAdd("M", Object(date));
+    if (reason.getLength() > 0)
+        vObj.dictAdd("Reason", Object(reason.copy()));
+    vObj.dictAdd("Contents", Object(objHexString, signature->copy()));
+    Object bObj(new Array(xref));
+    // reserve space in byte range for maximum number of bytes
+    bObj.arrayAdd(Object(static_cast<long long>(0LL)));
+    bObj.arrayAdd(Object(static_cast<long long>(9999999999LL)));
+    bObj.arrayAdd(Object(static_cast<long long>(9999999999LL)));
+    bObj.arrayAdd(Object(static_cast<long long>(9999999999LL)));
+    vObj.dictAdd("ByteRange", bObj.copy());
+    obj.dictSet("V", Object(vRef));
+    xref->setModifiedObject(&obj, ref);
+    return true;
+}
+
 std::vector<Goffset> FormWidgetSignature::getSignedRangeBounds() const
 {
     return static_cast<FormFieldSignature *>(field)->getSignedRangeBounds();
@@ -500,7 +741,8 @@ GooString *FormWidgetSignature::getCheckedSignature(Goffset *checkedFileSize)
 
 void FormWidgetSignature::updateWidgetAppearance()
 {
-    // Unimplemented
+    if (widget)
+        widget->updateAppearanceStream();
 }
 
 //========================================================================
@@ -1675,10 +1917,8 @@ void FormFieldChoice::reset(const std::vector<std::string> &excludedFields)
 // FormFieldSignature
 //------------------------------------------------------------------------
 FormFieldSignature::FormFieldSignature(PDFDoc *docA, Object &&dict, const Ref refA, FormField *parentA, std::set<int> *usedParents)
-    : FormField(docA, std::move(dict), refA, parentA, usedParents, formSignature), signature_type(unknown_signature_type), signature(nullptr), signature_info(nullptr)
+    : FormField(docA, std::move(dict), refA, parentA, usedParents, formSignature), signature_type(unknown_signature_type), signature(nullptr)
 {
-    signature = nullptr;
-
     signature_info = new SignatureInfo();
     parseInfo();
 }
@@ -1689,6 +1929,27 @@ FormFieldSignature::~FormFieldSignature()
     delete signature;
 }
 
+void FormFieldSignature::setSignature(const GooString &sig)
+{
+    delete signature;
+    signature = sig.copy();
+}
+
+const GooString &FormFieldSignature::getCustomAppearanceContent() const
+{
+    return customAppearanceContent;
+}
+
+void FormFieldSignature::setCustomAppearanceContent(const GooString &s)
+{
+    customAppearanceContent = GooString(s.toStr());
+}
+
+void FormFieldSignature::setCertificateInfo(std::unique_ptr<X509CertificateInfo> &certInfo)
+{
+    certificate_info.swap(certInfo);
+}
+
 void FormFieldSignature::parseInfo()
 {
     if (!obj.isDict())
@@ -1765,6 +2026,11 @@ FormSignatureType FormWidgetSignature::signatureType() const
     return static_cast<FormFieldSignature *>(field)->getSignatureType();
 }
 
+void FormWidgetSignature::setSignatureType(FormSignatureType fst)
+{
+    static_cast<FormFieldSignature *>(field)->setSignatureType(fst);
+}
+
 SignatureInfo *FormFieldSignature::validateSignature(bool doVerifyCert, bool forceRevalidation, time_t validationTime)
 {
 #ifdef ENABLE_NSS3
